@@ -5,13 +5,17 @@ from typing import Optional, List, Any, Iterable, Dict, Sequence, Union
 import logging
 
 from .core.values.value_mapping import get_map_helper
+from .core.type_mapper.data_model.base_data_model import TypeMapping
+from .core.type_mapper.modules.type_mapping_loader.type_mapping_loader import TypeMappingLoader
+from .core.type_mapper.modules.type_mapping_loader.type_mapping_loader_interface import TypeMappingLoaderInterface
 from .core.kb_sync.kb_iterator_config import KBIteratorConfig
 from .core.kb_sync.object_time_interval import ObjectTimeInterval
 from .tdm_builder.tdm_builder import AbstractTdmBuilder
 from .providers.gql_providers import AbstractGQLClient
 from .schema.api_schema import Query, Mutation, StoryPagination, ConceptLinkPagination, ConceptFactPagination, \
     ConceptPropertyPagination, SortDirection, ConceptSorting, DocumentSorting, DocumentGrouping, TimestampInterval, \
-    PerformSynchronously
+    PerformSynchronously, ConceptPropertyTypeSorting, ConceptTypeSorting, ConceptLinkTypeSorting, \
+    ConceptPropertyValueTypeSorting, CompositePropertyTypeSorting
 from .schema.api_schema import \
     ValueInput, IntValueInput, StringValueInput, DateTimeValue, StringLocaleValue, StringValue, LinkValue, \
     DoubleValue, IntValue, ConceptPropertyFilterSettings, ConceptLinkFilterSettings, ExtraSettings, Story, Document,\
@@ -49,21 +53,22 @@ def check_utils_gql_client(f):
 
 class TalismanAPIAdapter:
     def __init__(
-        self, gql_client: Optional[AbstractGQLClient], types: Dict, tdm_builder: AbstractTdmBuilder = None,
+        self, gql_client: Optional[AbstractGQLClient], type_mapping: Optional[Union[str, dict, TypeMapping]] = None,
+        type_mapping_loader: Optional[TypeMappingLoaderInterface] = None, tdm_builder: AbstractTdmBuilder = None,
         utils_gql_client: Optional[AbstractGQLClient] = None, kb_iterator_config: KBIteratorConfig = None,
         limit: int = 100, perform_synchronously: bool = True
     ) -> None:
         self._gql_client = gql_client
         self._utils_gql_client = utils_gql_client
-        self._types = types
-        self._cache = {}
+        self._type_mapping_loader = type_mapping_loader if type_mapping_loader else TypeMappingLoader(logger)
+        self._type_mapping = self._type_mapping_loader.load_type_mapping(type_mapping)
         self._limit = limit
         self._perform_synchronously = perform_synchronously
 
         self.document_fields_truncated = ('id', 'external_url')
         self.document_fields = (
             'id', 'title', 'external_url', 'publication_author', 'publication_date', 'internal_url', 'markers',
-            'system_registration_date', 'system_update_date', 'notes', 'access_level', 'trust_level'
+            'system_registration_date', 'system_update_date', 'notes', 'access_level', 'trust_level', 'uuid'
         )
         self.document_fields_extended = (
             'id', 'title', 'publication_author', 'publication_date', 'external_url',
@@ -1118,6 +1123,43 @@ class TalismanAPIAdapter:
 
         return res.delete_concept_property.is_success
 
+    def get_all_concept_types(
+        self, filter_settings: ConceptTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: ConceptTypeSorting = 'id'
+    ) -> Iterable[ConceptType]:
+        current_step = 0
+        while True:
+            concept_types = self.get_concept_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_types) < 1:
+                break
+            current_step += self.limit
+            for concept_type in concept_types:
+                yield concept_type
+
+    def get_concept_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: ConceptTypeSorting = 'id'
+    ) -> Sequence[ConceptType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = ConceptTypeFilterSettings()
+
+        op = Operation(Query)
+        pct: ConceptTypePagination = op.pagination_concept_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        pct.list_concept_type().__fields__(*self.concept_type_fields)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_concept_type.list_concept_type
+
     def get_concept_types_by_name(self, name: str) -> Sequence[ConceptType]:
         op = Operation(Query)
         ctp: ConceptTypePagination = op.pagination_concept_type(
@@ -1147,22 +1189,55 @@ class TalismanAPIAdapter:
         return res.concept_type
 
     def get_concept_type(self, concept_type_code: str) -> Optional[ConceptType]:
-        concept_type: Optional[ConceptType] = self._cache.get(f'concept_type_{concept_type_code}', None)
+        concept_type = self._type_mapping.get_concept_type(concept_type_code)
         if concept_type:
             return concept_type
-        concept_type_name = self._types['concepts_types_mapping'][concept_type_code]['name']
-        types = self.get_concept_types_by_name(concept_type_name)
-        if not types:
-            return None
-        concept_type = None
-        for t in types:
-            if t.name == concept_type_name:
-                concept_type = t
+
+        concept_type_name = self._type_mapping.get_concept_type_name(concept_type_code)
+        concept_types = self.get_concept_types_by_name(concept_type_name)
+        for concept_type in concept_types:
+            if concept_type.name == concept_type_name:
+                self._type_mapping.add_concept_type(concept_type_code, concept_type)
+                return concept_type
+
+    def get_all_concept_property_types(
+        self, filter_settings: ConceptPropertyTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: ConceptPropertyTypeSorting = 'name'
+    ) -> Iterable[ConceptPropertyType]:
+        current_step = 0
+        while True:
+            concept_property_types = self.get_concept_property_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_property_types) < 1:
                 break
-        if not concept_type:
-            return None
-        self._cache[f'concept_type_{concept_type_code}'] = concept_type
-        return concept_type
+            current_step += self.limit
+            for concept_property_type in concept_property_types:
+                yield concept_property_type
+
+    def get_concept_property_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptPropertyTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: ConceptPropertyTypeSorting = 'name'
+    ) -> Sequence[ConceptPropertyType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = ConceptPropertyTypeFilterSettings()
+
+        op = Operation(Query)
+        pcpt: ConceptPropertyTypePagination = op.pagination_concept_property_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        lcpt = pcpt.list_concept_property_type()
+        lcpt.__fields__(*self.concept_property_type_fields)
+        self._configure_property_value_type_fields(lcpt.value_type, True)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_concept_property_type.list_concept_property_type
 
     def get_concept_properties_types_by_name(
             self, concept_type_id: str, prop_name: str
@@ -1182,24 +1257,99 @@ class TalismanAPIAdapter:
         res = op + res
         return res.pagination_concept_property_type.list_concept_property_type
 
+    def get_all_concept_composite_property_types(
+        self, filter_settings: CompositePropertyTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: CompositePropertyTypeSorting = 'name'
+    ) -> Iterable[ConceptPropertyType]:
+        current_step = 0
+        while True:
+            concept_composite_property_types = self.get_concept_composite_property_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_composite_property_types) < 1:
+                break
+            current_step += self.limit
+            for concept_composite_property_type in concept_composite_property_types:
+                yield concept_composite_property_type
+
+    def get_concept_composite_property_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: CompositePropertyTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: CompositePropertyTypeSorting = 'name'
+    ) -> Sequence[ConceptPropertyType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = CompositePropertyTypeFilterSettings()
+
+        op = Operation(Query)
+        pccpt: ConceptPropertyTypePagination = op.pagination_composite_concept_property_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        lcpt = pccpt.list_concept_property_type()
+        lcpt.__fields__(*self.concept_property_type_fields)
+        self._configure_property_value_type_fields(lcpt.value_type, True)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_composite_concept_property_type.list_concept_property_type
+
     def get_concept_property_type(
         self, concept_type_code: str, property_type_code: str
     ) -> Optional[ConceptPropertyType]:
+        property_type = self._type_mapping.get_concept_property_type(concept_type_code, property_type_code)
+        if property_type:
+            return property_type
 
-        concept_type: ConceptType = self.get_concept_type(concept_type_code)
+        concept_type = self.get_concept_type(concept_type_code)
         if not concept_type:
             raise Exception('Cannot get concept property type: no concept type id')
 
-        prop_type = self._cache.get(f'concept_property_type_{concept_type_code}_{property_type_code}', None)
-        if prop_type:
-            return prop_type
-        prop_type_name = self._types['concepts_types_mapping'][concept_type_code]['properties'][property_type_code]
-        types = self.get_concept_properties_types_by_name(concept_type.id, prop_type_name)
-        for prop_type in types:
-            if prop_type.name == prop_type_name:
-                self._cache[f'concept_property_type_{concept_type_code}_{property_type_code}'] = prop_type
-                return prop_type
-        return None
+        property_type_name = self._type_mapping.get_concept_property_type_name(concept_type_code, property_type_code)
+        property_types = self.get_concept_properties_types_by_name(concept_type.id, property_type_name)
+        for property_type in property_types:
+            if property_type.name == property_type_name:
+                self._type_mapping.add_concept_property_type(concept_type_code, property_type_code, property_type)
+                return property_type
+
+    def get_all_concept_property_value_types(
+        self, filter_settings: ConceptPropertyValueTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: ConceptPropertyValueTypeSorting = 'id'
+    ) -> Iterable[ConceptPropertyValueType]:
+        current_step = 0
+        while True:
+            concept_property_value_types = self.get_concept_property_value_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_property_value_types) < 1:
+                break
+            current_step += self.limit
+            for concept_property_value_type in concept_property_value_types:
+                yield concept_property_value_type
+
+    def get_concept_property_value_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptPropertyValueTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: ConceptPropertyValueTypeSorting = 'id'
+    ) -> Sequence[ConceptPropertyValueType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = ConceptPropertyValueTypeFilterSettings()
+
+        op = Operation(Query)
+        pcpvt: ConceptPropertyValueTypePagination = op.pagination_concept_property_value_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        pcpvt.list_concept_property_value_type().__fields__(*self.cpvt_fields)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_concept_property_value_type.list_concept_property_value_type
 
     def get_concept_property_value_types_by_name(
         self, prop_value_type_name: str, limit: int = 20, offset: int = 0
@@ -1220,25 +1370,55 @@ class TalismanAPIAdapter:
     def get_concept_property_value_type(
         self, concept_property_value_type_code: str
     ) -> Optional[ConceptPropertyValueType]:
+        property_value_type = self._type_mapping.get_concept_property_value_type(concept_property_value_type_code)
+        if property_value_type:
+            return property_value_type
 
-        concept_property_value_type: Optional[ConceptPropertyValueType] = self._cache.get(
-            f'concept_property_value_type_{concept_property_value_type_code}', None
-        )
-        if concept_property_value_type:
-            return concept_property_value_type
-        concept_property_value_type_name = self._types['value_types_mapping'][concept_property_value_type_code]
-        types = self.get_concept_property_value_types_by_name(concept_property_value_type_name)
-        if not types:
-            return None
-        concept_property_value_type = None
-        for t in types:
-            if t.name == concept_property_value_type_name:
-                concept_property_value_type = t
+        value_type_name = self._type_mapping.get_concept_property_value_type_name(concept_property_value_type_code)
+        value_types = self.get_concept_property_value_types_by_name(value_type_name)
+        for value_type in value_types:
+            if value_type.name == value_type_name:
+                self._type_mapping.add_concept_property_value_type(concept_property_value_type_code, value_type)
+                return value_type
+
+    def get_all_concept_link_property_types(
+        self, filter_settings: ConceptPropertyTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: ConceptPropertyTypeSorting = 'name'
+    ) -> Iterable[ConceptPropertyType]:
+        current_step = 0
+        while True:
+            concept_link_property_types = self.get_link_property_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_link_property_types) < 1:
                 break
-        if not concept_property_value_type:
-            return None
-        self._cache[f'concept_property_value_type_{concept_property_value_type_code}'] = concept_property_value_type
-        return concept_property_value_type
+            current_step += self.limit
+            for concept_link_property_type in concept_link_property_types:
+                yield concept_link_property_type
+
+    def get_link_property_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptPropertyTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: ConceptPropertyTypeSorting = 'name'
+    ) -> Sequence[ConceptPropertyType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = ConceptPropertyTypeFilterSettings()
+
+        op = Operation(Query)
+        pclpt: ConceptPropertyTypePagination = op.pagination_concept_link_property_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        lcpt = pclpt.list_concept_property_type()
+        lcpt.__fields__(*self.concept_property_type_fields)
+        self._configure_property_value_type_fields(lcpt.value_type, True)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_concept_link_property_type.list_concept_property_type
 
     def get_link_properties_types_by_name(self, link_type_id: str, prop_name: str) -> Sequence[ConceptPropertyType]:
         op = Operation(Query)
@@ -1255,6 +1435,45 @@ class TalismanAPIAdapter:
         res = op + res
         return res.pagination_concept_link_property_type.list_concept_property_type
 
+    def get_all_concept_link_composite_property_types(
+        self, filter_settings: CompositePropertyTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: CompositePropertyTypeSorting = 'name'
+    ) -> Iterable[ConceptPropertyType]:
+        current_step = 0
+        while True:
+            concept_link_composite_property_types = self.get_link_composite_property_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_link_composite_property_types) < 1:
+                break
+            current_step += self.limit
+            for concept_link_composite_property_type in concept_link_composite_property_types:
+                yield concept_link_composite_property_type
+
+    def get_link_composite_property_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: CompositePropertyTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: CompositePropertyTypeSorting = 'name'
+    ) -> Sequence[ConceptPropertyType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = CompositePropertyTypeFilterSettings()
+
+        op = Operation(Query)
+        pclpt: ConceptPropertyTypePagination = op.pagination_composite_link_property_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        lcpt = pclpt.list_concept_property_type()
+        lcpt.__fields__(*self.concept_property_type_fields)
+        self._configure_property_value_type_fields(lcpt.value_type, True)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_composite_link_property_type.list_concept_property_type
+
     def get_composite_link_properties_types_by_name(
         self, link_type_id: str, prop_name: str
     ) -> Sequence[ConceptPropertyType]:
@@ -1273,36 +1492,36 @@ class TalismanAPIAdapter:
         return res.pagination_composite_link_property_type.list_concept_property_type
 
     def get_link_property_type(
-        self, link_type_code: str, link_type: ConceptLinkType,
-        property_type_code: str, is_composite: Optional[bool] = False
+        self, link_type_code: str, link_type: ConceptLinkType, property_type_code: str
     ) -> Optional[ConceptPropertyType]:
-        # link_type = self.get_link_type(link_type_code)
-        # if not link_type:
-        #     raise Exception('Cannot get link property type: no link type id')
+        property_type = self._type_mapping.get_concept_link_property_type(link_type_code, property_type_code)
+        if property_type:
+            return property_type
 
-        prop_type: Optional[ConceptPropertyType] = self._cache.get(
-            f'link_property_type_{link_type_code}_{property_type_code}', None
+        property_type_name = self._type_mapping.get_concept_link_property_type_name(link_type_code, property_type_code)
+        property_types = self.get_link_properties_types_by_name(link_type.id, property_type_name)
+        for property_type in property_types:
+            if property_type.name == property_type_name:
+                self._type_mapping.add_concept_link_property_type(link_type_code, property_type_code, property_type)
+                return property_type
+
+    def get_link_composite_property_type(
+        self, link_type_code: str, link_type: ConceptLinkType, property_type_code: str
+    ) -> Optional[ConceptPropertyType]:
+        property_type = self._type_mapping.get_concept_link_composite_property_type(link_type_code, property_type_code)
+        if property_type:
+            return property_type
+
+        property_type_name = self._type_mapping.get_concept_link_composite_property_type_name(
+            link_type_code, property_type_code
         )
-        if prop_type:
-            return prop_type
-        link_type_mapping = None
-        for r in self._types['relations_types_mapping']:
-            if r['old_relation_type'] == link_type_code:
-                link_type_mapping = r
-                break
-        if not link_type_mapping:
-            raise Exception('Cannot get link mapping object')
-        if is_composite:
-            prop_type_name = link_type_mapping['composite_properties'][property_type_code]['name']
-            types = self.get_composite_link_properties_types_by_name(link_type.id, prop_type_name)
-        else:
-            prop_type_name = link_type_mapping['properties'][property_type_code]
-            types = self.get_link_properties_types_by_name(link_type.id, prop_type_name)
-        for prop_type in types:
-            if prop_type.name == prop_type_name:
-                self._cache[f'link_property_type_{link_type_code}_{property_type_code}'] = prop_type
-                return prop_type
-        return None
+        property_types = self.get_composite_link_properties_types_by_name(link_type.id, property_type_name)
+        for property_type in property_types:
+            if property_type.name == property_type_name:
+                self._type_mapping.add_concept_link_composite_property_type(
+                    link_type_code, property_type_code, property_type
+                )
+                return property_type
 
     def add_property_by_id(
         self, id: str, type_id: str, value: Any, is_main: bool, value_type: str,
@@ -1473,16 +1692,15 @@ class TalismanAPIAdapter:
         component_value_types: List[CompositePropertyValueType],
         link_composite_property_code: str, link_code: str, is_main: bool, perform_synchronously: Optional[bool] = None
     ) -> ConceptProperty:
-        for relation in self._types['relations_types_mapping']:
-            if relation['old_relation_type'] != link_code:
-                continue
-            component_values = relation['composite_properties'][link_composite_property_code]['component_values']
-            components_type_mapping = self._get_components_mapping(component_values, component_value_types)
-            break
+        components_type_mapping: Dict[str, CompositePropertyValueType] = {}
+        for relation in self._type_mapping.relations_types_mapping:
+            if relation.old_relation_type == link_code:
+                component_values = relation.composite_properties[link_composite_property_code].component_values
+                components_type_mapping = self._get_components_mapping(component_values, component_value_types)
         value_input = self._get_value_input(values, components_type_mapping)
         perform_synchronously = self.get_perform_synchronously_value(perform_synchronously)
-        op = Operation(Mutation)
 
+        op = Operation(Mutation)
         aclp = op.add_concept_link_property(
             performance_control=PerformSynchronously(
                 perform_synchronously=perform_synchronously
@@ -1506,11 +1724,12 @@ class TalismanAPIAdapter:
         component_value_types: List[CompositePropertyValueType], concept_type_code: str, property_type_code: str,
         perform_synchronously: Optional[bool] = None
     ) -> ConceptProperty:
-        concept_composite_properties = self._types['concepts_types_mapping'][concept_type_code]['composite_properties']
-        concept_component_values = concept_composite_properties[property_type_code]['component_values']
+        concept_composite_properties = self._type_mapping.concepts_types_mapping[concept_type_code].composite_properties
+        concept_component_values = concept_composite_properties[property_type_code].component_values
         components_type_mapping = self._get_components_mapping(concept_component_values, component_value_types)
         value_input = self._get_value_input(values, components_type_mapping)
         perform_synchronously = self.get_perform_synchronously_value(perform_synchronously)
+
         op = Operation(Mutation)
         acp = op.add_concept_property(
             performance_control=PerformSynchronously(
@@ -1534,12 +1753,12 @@ class TalismanAPIAdapter:
         self, link_id: str, link_type_code: str, link_type: ConceptLinkType, property_type_code: str, value: Any,
         is_composite: Optional[bool] = False, is_main: bool = False, perform_synchronously: Optional[bool] = None
     ) -> ConceptProperty:
-
-        property_type: ConceptPropertyType = self.get_link_property_type(
-            link_type_code, link_type, property_type_code, is_composite
-        )
+        property_type = self.get_link_composite_property_type(
+            link_type_code, link_type, property_type_code
+        ) if is_composite else self.get_link_property_type(link_type_code, link_type, property_type_code)
         if not property_type:
             raise Exception('Cannot add property: no property type id')
+
         if is_composite:
             link_property = self.add_link_composite_property_by_id(
                 link_id, property_type.id, value, property_type.value_type.component_value_types,
@@ -1556,6 +1775,46 @@ class TalismanAPIAdapter:
 
         return link_property
 
+    def get_all_concept_link_types(
+        self, filter_settings: ConceptLinkTypeFilterSettings = None, direction: SortDirection = 'ascending',
+        sort_field: ConceptLinkTypeSorting = 'id'
+    ) -> Iterable[ConceptLinkType]:
+        current_step = 0
+        while True:
+            concept_link_types = self.get_concept_link_types(
+                skip=current_step, take=self.limit, filter_settings=filter_settings,
+                direction=direction, sort_field=sort_field
+            )
+            if len(concept_link_types) < 1:
+                break
+            current_step += self.limit
+            for concept_link_type in concept_link_types:
+                yield concept_link_type
+
+    def get_concept_link_types(
+        self, skip: int = 0, take: Optional[int] = None, filter_settings: ConceptLinkTypeFilterSettings = None,
+        direction: SortDirection = 'ascending', sort_field: ConceptLinkTypeSorting = 'id'
+    ) -> Sequence[ConceptLinkType]:
+        take = self.get_take_value(take)
+        if not filter_settings:
+            filter_settings = ConceptLinkTypeFilterSettings()
+
+        op = Operation(Query)
+        pclt: ConceptLinkTypePagination = op.pagination_concept_link_type(
+            direction=direction,
+            filter_settings=filter_settings,
+            limit=take,
+            offset=skip,
+            sort_field=sort_field
+        )
+        lclt = pclt.list_concept_link_type()
+        lclt.__fields__(*self.concept_link_type_fields)
+        lclt.concept_from_type().__fields__(*self.concept_type_fields)
+        lclt.concept_to_type().__fields__(*self.concept_type_fields)
+        res = self._gql_client.execute(op)
+        res = op + res
+        return res.pagination_concept_link_type.list_concept_link_type
+
     def get_concept_link_type_by_name(
         self, link_name: str, from_type_id: str, to_type_id: str, limit: int = 20
     ) -> Sequence[ConceptLinkType]:
@@ -1569,7 +1828,10 @@ class TalismanAPIAdapter:
             ),
             limit=limit
         )
-        pclt.list_concept_link_type().__fields__(*self.concept_link_type_fields)
+        lclt = pclt.list_concept_link_type()
+        lclt.__fields__(*self.concept_link_type_fields)
+        lclt.concept_from_type().__fields__(*self.concept_type_fields)
+        lclt.concept_to_type().__fields__(*self.concept_type_fields)
         res = self._gql_client.execute(op)
         res = op + res
         return res.pagination_concept_link_type.list_concept_link_type
@@ -1577,26 +1839,18 @@ class TalismanAPIAdapter:
     def get_link_type(
         self, concept_from_type_code: str, concept_to_type_code: str, link_type_code: str
     ) -> Optional[ConceptLinkType]:
-
-        link_type = self._cache.get(f'concept_link_type_{concept_from_type_code}_{link_type_code}', None)
+        link_type = self._type_mapping.get_concept_link_type(link_type_code)
         if link_type:
             return link_type
-        from_type = self.get_concept_type(concept_from_type_code)
-        to_type = self.get_concept_type(concept_to_type_code)
-        if not from_type or not to_type:
-            raise Exception('Cannot get link type: no concept type id')
-        link_type_name = None
-        for r in self._types['relations_types_mapping']:
-            if r['old_relation_type'] == link_type_code:
-                link_type_name = r['new_relation_type']
-        if not link_type_name:
-            raise Exception('Cannot get link type name')
-        types = self.get_concept_link_type_by_name(link_type_name, from_type.id, to_type.id)
-        for link_type in types:
+
+        concept_from_type = self.get_concept_type(concept_from_type_code)
+        concept_to_type = self.get_concept_type(concept_to_type_code)
+        link_type_name = self._type_mapping.get_concept_link_type_name(link_type_code)
+        link_types = self.get_concept_link_type_by_name(link_type_name, concept_from_type.id, concept_to_type.id)
+        for link_type in link_types:
             if link_type.name == link_type_name:
-                self._cache[f'concept_link_type_{concept_from_type_code}_{link_type_code}'] = link_type
+                self._type_mapping.add_concept_link_type(link_type_code, link_type)
                 return link_type
-        return None
 
     def add_relation_by_id(
         self, from_id: str, to_id: str, link_type_id: str, perform_synchronously: Optional[bool] = None
@@ -1789,6 +2043,16 @@ class TalismanAPIAdapter:
 
         return res.get_or_add_concept
 
+    @check_utils_gql_client
+    def get_tdm(self, doc_id: str):
+        op = Operation(uas.Query)
+        op.tdm(
+            id = doc_id
+        )
+        res = self._utils_gql_client.execute(op)
+        res = op + res
+        return res
+
     # endregion
 
     # region tcontroller methods
